/*
 * author Joerg Hansmann<info@jhansmann.de>
 * author Andreas Robinson<sleeperse@users.sourceforge.net>
 * author Moritz v. Buttlar<moritzvb@users.sourceforge.net>
 * author Jarek Foltynski <http://www.foltynski.net>
 * author Reiner Muench <moosec@gmx.de>	24.Sep.2005
 * 
 * Firmware v21.2 supports ModularEEG and MonolithEEG
 * changes for MonolithEEG(*1) and ModularEEG(*2)
 * 256/300/512/600/1024Hz/1200Hz sample rate 	*1+2   	
 * 57k6/115k2/230k4/460k8 Baud uart speed      	*1+2
 * 4 channel sampling   (modEEG 6 channels)     *1
 * Online/offline key                           *1
 * status led                                   *1
 * removed functions out(), in() , sbi() and cbi() - those are not recommended
 * in new code - see WinAVR documentation for more details
 * 
 * ModularEEG/MonolithEEG firmware based on p21 by Jarek Foltynski
 *
 * Based on the RS232EEG firmware v0.03 by Joerg Hansmann.
 * Compiles with AVR-GCC.
 *
 * FIXME: Electric Guru by Rob Sacks expects data in channel 0 and 1, but
 * modularEEG delivers data in channel 0 and 3.
 * In order to work with Electric Guru, channel 3 is copied to channel 1
 * in function SIGNAL(SIG_ADC). However, this is undesirable for other
 * with more than two channels.
 *
 * Program flow:
 *
 * 256Hz timer expires => SIGNAL(SIG_OVERFLOW0)	( base timing in both systems )
 *
 * SIGNAL(SIG_OVERFLOW0) enables the ADC
 *
 * Repeat 6 times: sampling starts and completes => SIGNAL(SIG_ADC)	( repeat 4 times in MonolithEEG )
 * 
 * SIGNAL(SIG_ADC) writes first byte to UART data register (UDR)
 *
 * Repeat 17 times: transmission begins, UDR empties => SIGNAL(SIG_UART_DATA)
 *
 * Start over from beginning.
  */

// ---------------------------------------------------------------------------
#define MonolithEEG			// change to comment for modEEG
// ---------------------------------------------------------------------------

#include <avr/io.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/pgmspace.h>

#define BIT(x)	(1<<x)

/*

////////// Packet format for Electric Guru compability //////////

// 17-byte packets are transmitted from the ModularEEG at 256Hz,
// using 1 start bit, 8 data bits, 1 stop bit, no parity, 57600 bits per second.

// Minimial transmission speed is 256Hz * sizeof(modeeg_packet) * 10 = 43520 bps.

struct modeeg_packet
{
uint8_tsync0;// = 0xa5
uint8_tsync1;// = 0x5a
uint8_tversion;// = 2
uint8_tcount;// packet counter. Increases by 1 each packet.
uint16_tdata[6];// 10-bit sample (= 0 - 1023) in big endian (Motorola) format.
uint8_tswitches;// State of PD5 to PD2, in bits 3 to 0.
};


// Note that data is transmitted in big-endian format.
// By this measure together with the unique pattern in sync0 and sync1 it is guaranteed, 
// that re-sync (i.e after disconnecting the data line) is always safe.

*/

/*         
        !!! HAVE ATTENTION TO SAMPLE RATE AND BAUD RATE RATIO !!!!

                Minimal baud rates for different packet types
             --------------------------------------------------------
                17 Byte Packet              5 Byte Packet 
                ElectricGuru                Bioera
             --------------------------------------------------------    
    Sample      Baud                        Baud
    rate        rate                        rate
    [Hz]        kbps                        kbps
    -----------------------------------------------------------------
     256        57.6						
     300
     512
     600
    1024
    1200       



*/




// ---------------------------------------------------------------------------
#define TESTFREQ	14			// default adjustment frequence in Hz @ 256Hz

#define	XTAL	7372800     // MonolithEEG & modEEG

#if defined (MonolithEEG)
#define BLINKTIME 127		    // 1s blink @ 256Hz
#define SAMPLECHANNELS 4

#else	// modEEG
#define SAMPLECHANNELS 6
#endif

//---------------------------------------------------------------
//	Port Definition	MonolithEEG
//---------------------------------------------------------------
// PORT B	MonolithEEG
#define IO_ICP        	BIT(0)  					// connected to PD6 via 220R resitor in MonolithEEG
#define	IO_OC1A        	BIT(1)                   // Test PWM
#define	IO_OC1B        	BIT(2)                   //
#define	IO_OC2        	BIT(3)                   // MOSI, OC2
#define	IO_MISO        	BIT(4)                   // MISO
#define	IO_SCK        	BIT(5)                   // SCK
//		        		BIT(6)  					// XTAL
//      			   	BIT(7)      				// XTAL

//#define TestToggle()   	( PORTB ^=  BIT(1) )		// adjustment signal in MonolithEEG
#define DDRB_CFG	( IO_SCK | IO_MISO | IO_OC1A )
#define PORTB_CFG	( IO_SCK | IO_MISO | IO_OC2 | IO_OC1B | IO_OC1A )

//---------------------------------------------------------------
//	Port C	MonolithEEG
#define	CH1		    	BIT(0)		               	// EEG Channel 1
#define	CH2		    	BIT(1)		               	// EEG Channel 2
#define	ACH1	    	BIT(2)		               	// AUX Channel 1
#define	ACH2	    	BIT(3)		               	// AUX Channel 2
#define	KEY		    	BIT(4)		               	// Key
#define	POWER_LED    	BIT(5)		               	// LED
//						BIT(5)						// RESET

#define PowerLedOff()  		( PORTC &= ~POWER_LED )
#define PowerLedOn()   		( PORTC |=  POWER_LED )
#define PowerLedToggle()   	( PORTC ^=  POWER_LED )
#define	KeyEEG()			(!(PINC & KEY))			// get key true=pressed

#define DDRC_CFG    ( POWER_LED )					// outputs
#define PORTC_CFG	( KEY )							// Pullups

//---------------------------------------------------------------
//	Port D	MonolithEEG
//#define   RxD         BIT(0)
#define	TxD         	BIT(1)
#define	IO_INT0        	BIT(2)
#define	IO_INT1        	BIT(3)
#define	IO_PD4	     	BIT(4)
#define	IO_PD5         	BIT(5)
#define	IO_PD6         	BIT(6)
#define	IO_PD7         	BIT(7)


//#define DDRD_CFG	( IO_PD7 | IO_PD6 )
#define DDRD_CFG	( IO_PD7 | IO_PD6 | IO_PD5 | IO_PD4 | IO_INT1 | IO_INT0 )
#define PORTD_CFG	( IO_PD7 | IO_PD6 | IO_PD5 | IO_PD4 | IO_INT1 | IO_INT0 | TxD )

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

#define NUMCHANNELS 6
#define HEADERLEN 4
#define PACKETLEN (NUMCHANNELS * 2 + HEADERLEN + 1)

#define SAMPFREQ 		256			// standard sample rate

//	Timer cycles
// 256 - Fosc/pre/SamplerRate + 0.5 ) = (513 - 2*Fosc/pre/SamplerRate)/2	include rounding +0.5 for better fit
#define TIMER0VAL_256	( 513 - 2 * XTAL / 256 /  256 ) / 2		// Precounter 1:256
#define TIMER0VAL_300	( 513 - 2 * XTAL / 256 /  300 ) / 2		// Precounter 1:256
#define TIMER0VAL_512	( 513 - 2 * XTAL / 256 /  512 ) / 2		// Precounter 1:256
#define TIMER0VAL_600	( 513 - 2 * XTAL / 256 /  600 ) / 2		// Precounter 1:256
#define TIMER0VAL_1024	( 513 - 2 * XTAL / 256 / 1024 ) / 2 	// Precounter 1:256
#define TIMER0VAL_1200	( 513 - 2 * XTAL / 256 / 1200 ) / 2 	// Precounter 1:256


#define	TIMER_N	6
const uint8_t TimerCyleTab[ TIMER_N ] PROGMEM = {
	TIMER0VAL_256,
	TIMER0VAL_300,
	TIMER0VAL_512,
   	TIMER0VAL_600,
   	TIMER0VAL_1024,
   	TIMER0VAL_1200,
};

uint8_t TimerTabIndex = 0;      // default 256Hz
// Fosc/pre/Baud-0.5 = (2*Fosc/pre/Baud-1)	include rounding +0.5 for better fit
#define BR_57600	( XTAL / 8 / 57600 - 1 ) / 2 		// Precounter 1:16 ( U2X = 0)
#define BR_115200	( XTAL / 8 / 115200 - 1 ) / 2 		// Precounter 1:16 ( U2X = 0)
#define BR_230400	( XTAL / 8 / 230400 - 1 ) / 2 		// Precounter 1:16 ( U2X = 0)
#define BR_460800	( XTAL / 8 / 460800 - 1 ) / 2 		// Precounter 1:16 ( U2X = 0)

#define	BAUD_N	4
const uint8_t BaudRateTab[ BAUD_N ] PROGMEM = {
	BR_57600,
	BR_115200,
	BR_230400,
	BR_460800,
};
uint8_t BaudRateTabIndex = 0;   // default 57600Baud

 /** The transmission packet */
volatile uint8_t TXBuf[2 * NUMCHANNELS + HEADERLEN + 1];

/** Next byte to read or write in the transmission packet. */

volatile uint8_t TXIndex;  

/** Current channel being sampled. */

volatile uint8_t CurrentCh;

volatile uint8_t KeyTimer;						
volatile uint8_t TestTimer;						
volatile uint8_t BlinkTimer;
volatile uint8_t TestFrequence;
volatile uint8_t runEEG;
volatile uint8_t TimerReload;

/* ------------------------ Protocol 21 add-on ------------------------------*/
//
//Packet has variable length, it depends on how many channels we want to receive
//Each channel can be set indywidually via backward communication
//
//Packet format
//1cccnnnn
//0pppaaaa
//0bbbbbbb
//....
//0pppaaaa
//0bbbbbbb
//
//
//ccc  - number of channels in packet, total packet length = 1 + (2 * ccc);
//nnnn - 4 bit control sequence number
//ppp  - channel id, 0-5 are A/D channels, 
//6 is for backward information (requested or error code)
//7 - not used now
//aaaa - if the ppp is 0-5, then the lower three bits of aaaa contains 3 highest bits of 10bit channel value, the highest bit of aaaa is set to 0
//       if the ppp is 6, then it contains id/selector of backward information       
//bbbbbbb - if the ppp is 0-5, then it contains 7 lower bits of 10bit value
//          if the ppp is 6, then it contains backward information value   
//
//

volatile uint8_t ActiveChannelsMatrix;  // bits: 1 - ch 0, 2 - ch 1, ..., 6 - ch 5, 7 - not used
volatile uint8_t ActiveChannelsNumber;  // how many channels are active, contains value between 1 and 7
volatile uint8_t RequestedInformation;  // shows that an information was requested
volatile uint8_t PackageCounter;// 4-bit sequencial counter
volatile uint8_t DataIndex;// used for packet creation
volatile uint8_t PackageLength;// total packet length (1 + 2 * channels_number)
volatile uint8_t ProtocolNumber;// contains numeric "2" (default) or "21"

void setBits(void);// Used for controlling state of portD bits
void initProtocol21(void);// Protocol 21 initializations
uint8_t GetRequestedInformationValue(void);

/* ------------------------ Protocol 21 end ----------------------------------*/

// ---------------------------------------------------------------------------
/** Sampling timer (timer 0) interrupt handler */

SIGNAL(SIG_OVERFLOW0)
{
    volatile uint8_t i;

	TCNT0 = TimerReload;				// Reset timer to get correct sampling frequency.

#if defined (MonolithEEG)		// only for MonolithEEG

	if( KeyEEG() ){					// check Key
		if( KeyTimer ){				// is debounce runing?
			KeyTimer--;				// decrement time
			if( !KeyTimer ){		// time out?
				runEEG ^= 1;		// toggle run/stop
				BlinkTimer = 0;     // reset blink timer
				if( !runEEG ){		// 
					PowerLedOff();
				}
			}
		}
	}
	else{
		KeyTimer = 20;				// set debounce time
	}
#endif

	if( runEEG ){					// enable/disable packets
	#if defined (MonolithEEG)		// only for MonolithEEG
		PowerLedOn();
	#endif
		if (ProtocolNumber == 21) {
			// jarek's new protocol
			PackageCounter = (PackageCounter + 1) & 0xF; 
			TXBuf[0] = 128 + PackageCounter;  // Prepare start byte
			PackageLength = 1 + (ActiveChannelsNumber * 2); 
			DataIndex = 1;
			if (RequestedInformation) {
				// Increase the packet length by 2 bytes that will contains the additional requested data
				PackageLength += 2;
				// Set value of the requested information
				TXBuf[PackageLength - 2] = (6 << 4) + (RequestedInformation & 0xF);
				TXBuf[PackageLength - 1] = GetRequestedInformationValue();
				RequestedInformation = 0;  // Reset the requested information id
				TXBuf[0] += (((ActiveChannelsNumber + 1) & 7) << 4);
			} else {
				TXBuf[0] += ((ActiveChannelsNumber & 7) << 4);
			}
			if (ActiveChannelsNumber > 0) {
				// Search for first active channel
				for (i = 0; i < ActiveChannelsNumber; i++) {
					if (ActiveChannelsMatrix & (1 << i)) {
						CurrentCh = i;
						break;
					}
				}
			} else {
				// There no active channels, even without any channel active we want to go to sample procedure, 
				// to keep the source code compatible with previous version
				CurrentCh = 0;
			}
		} else {
			// Here is the default protocol
			PackageLength = PACKETLEN;
			CurrentCh = 0;

			// Write header and footer:

			// Increase packet counter (fourth byte in header)

			TXBuf[3]++;

			//Get state of switches on PD2..5, if any (last byte in packet).

			TXBuf[2 * NUMCHANNELS + HEADERLEN] = ( PIND >> 2 ) & 0x0F;

		}

		ADMUX = CurrentCh;		// Prepare conversion on channel.
		UCSRB &= ~BIT(UDRIE);  	// Ensure UART IRQ's are disabled.
		ADCSR |=  BIT(ADIF);    // Reset any pending ADC interrupts
		ADCSR |=  BIT(ADIE);    // Enable ADC interrupts.

//		ADCSR |=  BIT(ADSC);    // Start Conversion
		//The ADC will start sampling automatically as soon
		//as sleep is executed in the main-loop.
	}
#if defined (MonolithEEG)		// only in MonolithEEG
	else{							// show offline with blinking led
		if( ++BlinkTimer > BLINKTIME ){	// LED blink
			BlinkTimer = 0;
			PowerLedToggle();
		}
	}
#endif

	setBits();
}

// ---------------------------------------------------------------------------
/** AD-conversion-complete interrupt handler. */

SIGNAL(SIG_ADC){
    volatile uint8_t i;

//	PORTB ^= BIT(5);				// generate square

    if (ProtocolNumber == 21) {
		if (ActiveChannelsMatrix & (1 << CurrentCh)) {
			// ok we have an active channel here, lets read it
			i = ADCL;
			TXBuf[DataIndex + 1] = i & 127;
			TXBuf[DataIndex] = (CurrentCh << 4) + (ADCH << 1) + (i >> 7);
			DataIndex += 2;
			// Find next active channel
			while (++CurrentCh < SAMPLECHANNELS) {
				if (ActiveChannelsMatrix & (1 << CurrentCh)) {
					// Ok, this channel is active
					break;
				}
			}
		} else {
			// The current ChanneNo is not active, no more to do here
			CurrentCh = NUMCHANNELS;
		}
	} else {
		i = 2 * CurrentCh + HEADERLEN;
		TXBuf[i+1] 	= ADCL;
		TXBuf[i] 	= ADCH;
		CurrentCh++;
	}

	if (CurrentCh < SAMPLECHANNELS){
		ADMUX = CurrentCh;	//Select the next channel.
			//The next sampling is started automatically.
	}
	else
	{
		ADMUX = 0;			//Prepare next conversion, on channel 0.

#if !defined (MonolithEEG)		// only in modEEG
		if (ProtocolNumber == 2) {
			//Electric guru kludge. Maps channel 3 to channel 1.
			TXBuf[HEADERLEN + 2] = TXBuf[HEADERLEN + 6];
			TXBuf[HEADERLEN + 3] = TXBuf[HEADERLEN + 7];
		}
#endif
		// Disable ADC interrupts to prevent further calls to SIG_ADC.
		ADCSR &= ~BIT(ADIE);

		// Hand over to SIG_UART_DATA, by starting
		// the UART transfer and enabling UDR IRQ's.

		UDR = TXBuf[0]; 
		UCSRB |= BIT(UDRIE);
		TXIndex = 1;
	}
//	ADCSR |=  BIT(ADSC);    // Start Conversion
}

// ---------------------------------------------------------------------------
/*** UART data transmission register-empty interrupt handler ***/

SIGNAL(SIG_UART_DATA)
{
    UDR = TXBuf[TXIndex];			//Send next byte
    TXIndex++;
    if (TXIndex == PackageLength)   //See if we're done with this packet
    {
        UCSRB &= ~BIT(UDRIE);		//Disable SIG_UART_DATA interrupts.
									//Next interrupt will be a SIG_OVERFLOW0.
    }
}

// ---------------------------------------------------------------------------
//#if !defined (MonolithEEG)			// include only in modEEG
/** Initialize PWM output (PB1 = 14Hz square wave signal) */
void pwm_init(void)
{
#if defined (__AVR_ATmega8__)
	OCR1A = 512;		//This causes the signal to have roughly 50% symmetry.
	TCCR1A = BIT(COM1A1) + BIT(WGM11) + BIT(WGM10);	//Set PWM-mode: non-inverted toggling, 10 bit counter.
	TCCR1B = BIT(CS12);	//Start the timer, set to count at clk / 256
#else	// AT90S4433
	OCR1 = 512;			//This causes the signal to have roughly 50% symmetry.
	TCCR1A = BIT(COM11) + BIT(PWM11) + BIT(PWM10);	//Set PWM-mode: non-inverted toggling, 10 bit counter.
	TCCR1B = BIT(CS12);	//Start the timer, set to count at clk / 256 = 28800Hz
#endif
    // Effective frequency is clk / 256 / 2046 = 14Hz.
}
//#endif

/* ------------------------ Protocol 21 Add-On ----------------------------------*/

volatile uint8_t commandByte = 0;// Command byte range is 128-255
volatile uint8_t dataByte = 128;// Data byte range is 0-127
volatile uint8_t portDBitIsOn[8]; //
volatile uint8_t portDBitBrightPeriod[8];  //
volatile uint8_t portDBitBrightOnTime[8];  //
volatile uint8_t portDBitBrightCounter[8]; //

// Contains commands ids (set in 7 lower bits of the command byte)
#define RECV_PORT_SET 1
#define RECV_PORTD_BIT_SET 2
#define RECV_PORTD_BIT_BRIGHT_ONTIME 3
#define RECV_PORTD_BIT_BRIGHT_PERIOD 4
#define RECV_CHANNEL_SET 5
#define RECV_GET_VINFO 6
#define RECV_SET_VINFO 7

// Here are values that can be set or requested, they are used with RECV_GET_VINFO and RECV_SET_VINFO
#define VINFO_PROTOCOL_NUMBER 1
#define VINFO_CHANNELS_MATRIX 2
#define VINFO_BAUD_RATE 3       // set baud rate (set only)
#define VINFO_PORTD 4
#define VINFO_PORTB 5
#define VINFO_SAMPLE_RATE 6     	// sample rate
#define VINFO_RUNEEG 7          	// start/stop EEG function                  0:stop, 1:run
#define VINFO_DEVICE 8          	// what kind of EEG is online (read only)   answer 0:modEEG, 1:MonolithEEG
#define VINFO_PROTOCOL_SUBNUMBER 9 	// view sub version of protocoll 21
//....
//#define  15// This max value for requested information

// ---------------------------------------------------------------------------
SIGNAL(SIG_UART_RECV)
{
	volatile uint8_t i;
	volatile uint8_t n;
	volatile uint8_t rxbuf;

	rxbuf = UDR;

	if (rxbuf > 127) {
		// This is command byte
		commandByte = rxbuf;// Set command value
		dataByte = 128;// Reset data value
		return;
	} else if (commandByte == 0) {
		// If command not yet set, disregard this
		return;
	}

	if (dataByte == 128) {
		// This is the first byte of data, lets wait for the next
		dataByte = rxbuf;
		return;
	}

	// Compare the lower 7 bits in the command byte
	switch (commandByte & 127) {
		case RECV_PORT_SET:
			// data1 - port selector
			// data2 - value of the port shifted 2 positions
			if (dataByte == 'B')
				PORTB = rxbuf;
			else 
				PORTD = rxbuf << 2;    
			break;
		case RECV_PORTD_BIT_BRIGHT_ONTIME:
			// data1 - bit number
			// data2 - value
			portDBitBrightOnTime[dataByte] = rxbuf;
			break;
		case RECV_PORTD_BIT_BRIGHT_PERIOD:
			// data1 - bit number
			// data2 - value
			portDBitBrightPeriod[dataByte] = rxbuf;
			break;
		case RECV_PORTD_BIT_SET:
			// data1 - bit number
			// data2 - 0 to reset or anything else to set the bit
			if (rxbuf) {
				// Set the bit
				portDBitIsOn[dataByte] = 1;
				portDBitBrightCounter[dataByte] = 0;
				PORTD = PIND | (1<<dataByte);
			} else {
				// Reset the bit
				portDBitIsOn[dataByte] = 0;
				portDBitBrightCounter[dataByte] = portDBitBrightOnTime[dataByte];
				PORTD = PIND & ~(1<<dataByte);
			}
			break;
		case RECV_GET_VINFO:
			// data1 - Variable information ID
			// data2 - not used
			RequestedInformation = dataByte;
			break;
		case RECV_SET_VINFO:
			// data1 - Variable information ID
			// data2 - value to set
			switch (dataByte) {
				case VINFO_CHANNELS_MATRIX:
					ActiveChannelsMatrix = rxbuf;
					n = 0;
					for (i = 0; i < 7; i++) {
						if (ActiveChannelsMatrix & (1 << i)) {
							n++;
						}
					}
					ActiveChannelsNumber = n;
					break;
				case VINFO_PROTOCOL_NUMBER:
					ProtocolNumber = rxbuf;
					if (ProtocolNumber == 2) {
						// Initialize again the table
						TXBuf[0] = 0xa5;        // Sync 0
						TXBuf[1] = 0x5a;        // Sync 1
						TXBuf[2] = 2;           // Protocol version
						TXBuf[3] = 0;           // Packet counter
						TXBuf[12]= 2;			// dummy channel 5
						TXBuf[13]= 0;           // dummy channel 5
						TXBuf[14]= 2;           // dummy channel 6
						TXBuf[15]= 0;           // dummy channel 6
					}
					break;
				case VINFO_BAUD_RATE:
                    if( rxbuf < BAUD_N ){      // limit index
                        BaudRateTabIndex = rxbuf;
                        #if defined (__AVR_ATmega8__)
                        UCSRB = 0;	// transmitter and receiver off
                        UBRRH = 0;
                        UBRRL = pgm_read_byte( &BaudRateTab[ BaudRateTabIndex ] );	// Set speed to 57600 bps
                        UCSRB = BIT(TXEN) | BIT(RXEN) | BIT(RXCIE);	// transmitter and receiver enable, receive interrupt on
                        #else	// AT90S4433
                        UCSRB = 0;	// transmitter and receiver off
                        UBRR = pgm_read_byte( &BaudRateTab[ BaudRateTabIndex ] );	// Set speed to 57600 bps
                        UCSRB = BIT(TXEN) + BIT(RXEN) + BIT(RXCIE);    // transmitter and receiver enable, receive interrupt on
                        #endif
                    }
					break;
                case VINFO_SAMPLE_RATE:
                    if( rxbuf < TIMER_N ){      // limit index
                        TimerTabIndex = rxbuf;
                        TimerReload =  pgm_read_byte( &TimerCyleTab[ TimerTabIndex ] );     // set Sample rate to 256Hz
                    }
                    break;
                case VINFO_RUNEEG:
                    runEEG = rxbuf&1;   // set run/stop
                    break;
				default:
					break;
			}
			break;
		case RECV_CHANNEL_SET:
			// data1 - channel number
			// data2 - 0 to reset, anything else to set the bit
			if (rxbuf) {
				// Set the bit
				if (ActiveChannelsMatrix & (1 << dataByte)) {
					// Channel is already set, do nothing
				} else {
					ActiveChannelsMatrix |= (1 << dataByte);
					ActiveChannelsNumber++;
				}
			} else {
				// Reset
				if (ActiveChannelsMatrix & (1 << dataByte)) {
					ActiveChannelsMatrix &= ~(1 << dataByte);
					ActiveChannelsNumber--;
				} else {
					// Channel is already not set, do nothing
				}
			}
			break;
		default:
			// Unknown command, clear everything and quit
			break;
	}

	commandByte = 0;

}

// ---------------------------------------------------------------------------
void initProtocol21() {
	volatile uint8_t i;

		for(i = 0; i < 8; i++) {
			portDBitIsOn[i] = 1;  // Turn on all bits
			portDBitBrightPeriod[i] = 0;
			portDBitBrightOnTime[i] = 0;
			portDBitBrightCounter[i]= 0;
		}

		ActiveChannelsMatrix = 127;// All channels are active by default
		ActiveChannelsNumber = 6;
		RequestedInformation = 0;// No information requested yet
		ProtocolNumber = 2;// ElecGuru protocol is default
		PackageLength = PACKETLEN;
		PackageCounter = 0;
		DataIndex = 0;
	}

// ---------------------------------------------------------------------------
#if defined (MonolithEEG)		// 
#define SHIFT   0           // in MonolithEEG
#else
#define SHIFT   2           // in modEEG
#endif

#define STARTCH (0+SHIFT)
#define LASTCH  (5+SHIFT)

void setBits(void) {
	volatile uint8_t i;

	// Set brightness of all 6 upper bits of port D
	for (i = STARTCH; i <= LASTCH; i++) {
		if (portDBitIsOn[i] && portDBitBrightPeriod[i] != 0) {
			// Control the bit brightness
			portDBitBrightCounter[i]++;
			if (portDBitBrightCounter[i] == portDBitBrightPeriod[i]) {
				// Reset counter
				portDBitBrightCounter[i] = 0;

				// Turn the bit on
				PORTD = PIND | BIT(i);
			}
			if (portDBitBrightCounter[i] == portDBitBrightOnTime[i]) {
				// Turn the bit off
				PORTD = PIND & ~BIT(i);
			}
		}
	}
}

// ---------------------------------------------------------------------------
uint8_t GetRequestedInformationValue(void) {
	uint8_t ret = 0;

	switch (RequestedInformation) {
		case VINFO_PROTOCOL_NUMBER:
			ret = ProtocolNumber;
			break;
		case VINFO_CHANNELS_MATRIX:
			ret = ActiveChannelsMatrix;
			break;
		case VINFO_BAUD_RATE:
			// not implemented
			break;
		case VINFO_PORTD:
			// pass only upped 6 bits of port D
			ret = PIND >> 2;
			break;
		case VINFO_PORTB:
			ret = PINB;
			break;
        case VINFO_SAMPLE_RATE:
               ret = TimerTabIndex;
            break;
        case VINFO_RUNEEG:
            ret = runEEG;   // view status
            break;
        case VINFO_DEVICE:
			#if defined (MonolithEEG)
			ret = 1;        // I'am a MonolithEEG
			#else	    
			ret = 0;        // I'am a modEEG
			#endif
            break;
		case VINFO_PROTOCOL_SUBNUMBER:
			ret = 2;        // Sub version 2
			break;
		default:
			break;
	}

	// Make sure, that the value has up to 7 bits
	return ret & 127;
}

/* ------------------------ Protocol 21 end ----------------------------------*/


// ---------------------------------------------------------------------------
int main( void ){
    //Write packet header and footer

	TXBuf[0] = 0xa5;        // Sync 0
	TXBuf[1] = 0x5a;        // Sync 1
	TXBuf[2] = 2;           // Protocol version
	TXBuf[3] = 0;           // Packet counter
	TXBuf[12]= 2;			// dummy channel 5 in MonolithEEG
	TXBuf[13]= 0;           // dummy channel 5 in MonolithEEG
	TXBuf[14]= 2;           // dummy channel 6 in MonolithEEG
	TXBuf[15]= 0;           // dummy channel 6 in MonolithEEG

    //Set up the ports.
#if defined ( MonolithEEG )
	PORTB = PORTB_CFG;
	PORTC = PORTC_CFG;
	PORTD = PORTD_CFG;

	DDRB  = DDRB_CFG;
	DDRC  = DDRC_CFG;
	DDRD  = DDRD_CFG;
    runEEG = 0;		// MonolithEEG is offline on reset
#else		        // modEEG
	DDRD  = 0xc2;
	DDRB  = 0x07;
	PORTD = 0xff;
	PORTB = 0xff;
	runEEG = 1;		// modEEG start sending packets immediately
#endif

    // Select sleep mode = idle.

#if defined (__AVR_ATmega8__)
	MCUCR = ( MCUCR | BIT(SE)) & (~BIT(SM0) | ~BIT(SM1) | ~BIT(SM2) );
#else	// AT90S4433
	MCUCR = ( MCUCR | BIT(SE)) & (~BIT(SM));
#endif

    // Initialize the ADC

    // Timings for sampling of one 10-bit AD-value:
    //
    // prescaler > ((XTAL / 200kHz) = 36.8 => 
    // prescaler = 64 (ADPS2 = 1, ADPS1 = 1, ADPS0 = 0)
    // ADCYCLE = XTAL / prescaler = 115200Hz or 8.68 us/cycle
    // 14 (single conversion) cycles = 121.5 us (8230 samples/sec)
    // 26 (1st conversion) cycles = 225.69 us

    ADMUX = 0;      // Select channel 0

	ADCSR = BIT(ADPS2) + BIT(ADPS1);      // 1:64 - 115200Hz ADC clock
	ADCSR |= BIT(ADIF);		// Reset any pending ADC interrupts
	ADCSR |= BIT(ADEN);      // Enable the ADC
                              
    // Initialize the UART  RX & TX on for bidirectional communication

	BaudRateTabIndex = 0;

#if defined (__AVR_ATmega8__)
	UBRRH = 0;
	UBRRL = pgm_read_byte( &BaudRateTab[ BaudRateTabIndex ] );	// Set speed to 57600 bps
	UCSRB = BIT(TXEN) | BIT(RXEN) | BIT(RXCIE);	// transmitter and receiver enable, receive interrupt on
#else	// AT90S4433
	UBRR = pgm_read_byte( &BaudRateTab[ BaudRateTabIndex ] );	// Set speed to 57600 bps
	UCSRB = BIT(TXEN) + BIT(RXEN) + BIT(RXCIE);    // transmitter and receiver enable, receive interrupt on
#endif

	// Initialize timer 0    
	TimerTabIndex = 0;
	TCNT0 = 0;						// Clear it.
	TimerReload =  pgm_read_byte( &TimerCyleTab[ TimerTabIndex ] );     // set Sample rate to 256Hz
	TCCR0 = 4;             		// Start it. Frequency = clk / 256
	TIMSK = BIT(TOIE0);     			// Enable the interrupts.

    // Initialize PWM (optional)
	pwm_init();                                                  																			
    initProtocol21();

    sei();

    // Now, we wait. This is an event-driven program, so nothing much
    // happens here.

    while (1){
        __asm__ __volatile__ ("sleep");
    }
}

// ---------------------------------------------------------------------------
// end
// ---------------------------------------------------------------------------



/*

Command,	set MSB in command byte

RECV_PORT_SET 					1		set/clear separate port bits
RECV_PORTD_BIT_SET 				2
RECV_PORTD_BIT_BRIGHT_ONTIME 	3
RECV_PORTD_BIT_BRIGHT_PERIOD 	4
RECV_CHANNEL_SET 				5		
RECV_GET_VINFO 					6       read parameter
RECV_SET_VINFO 					7		write parameter

// Here are values that can be set or requested, they are used with RECV_GET_VINFO and RECV_SET_VINFO
#define VINFO_PROTOCOL_NUMBER 1
#define VINFO_CHANNELS_MATRIX 2
#define VINFO_BAUD_RATE 3       // set baud rate (set only)
#define VINFO_PORTD 4
#define VINFO_PORTB 5
#define VINFO_SAMPLE_RATE 6     	// sample rate
#define VINFO_RUNEEG 7          	// start/stop EEG function                  0:stop, 1:run
#define VINFO_DEVICE 8          	// what kind of EEG is online (read only)   answer 0:modEEG, 1:MonolithEEG
#define VINFO_PROTOCOL_SUBNUMBER 9 	// view sub version of protocoll 21

*/
